
/* Copyright by A.N. Yzelman, Dept. of Computer Science, K.U.Leuven, 2011.
 * Released under GPL, see the file LICENSE.txt
 */

#include <iostream>
#include <vector>
#include <map>
#include <pthread.h>
#include <numa.h>

//#include "hwloc.h"

#include "SparseMatrix.hpp"
#include "Matrix2HilbertCoordinates.hpp"
#include "FBICRS.hpp"
#include "MachineInfo.hpp"
#include "BigInt.hpp"

#ifndef _H_RDBHILBERT
#define _H_RDBHILBERT

/*
 * When defined, thread 0 will use the global y vector for its local
 * computations. This introduces extra work in the form of one sync,
 * and the amount of available processors for the parallel collect.
 * The advantage is less data replication of the output vector.
 *
 * The synchronisation cannot be prevented. Using all processors in
 * the collect code is possible but has not been programmed currently.
 */
#define RDBH_GLOBAL_Y

/*
 * When defined, RDBHilbert will not collect output results in the
 * global output vector passed to this library. Used for timing
 * the true SpMV speeds only.
 */
//#define RDBH_NO_COLLECT

/** Shared data for RDBHilbert threads. */
template< typename T >
class RDB_shared_data {

	public:

		unsigned long int id;

		unsigned long int P;

		/** 0 undef, 1 init, 2 zax, 3 zxa, 4 exit */
		unsigned char mode;

		/** how many times to repeat the operation set in `mode' (above, only for 2 and 3) */
		unsigned long int repeat;

		std::vector< Triplet< T > > *original;

		/** Will store rowsums */
		unsigned long int *nzb;

		/** Will store local timing */
		double time;

		pthread_mutex_t* mutex;

		pthread_cond_t*  cond;

		pthread_mutex_t* end_mutex;

		pthread_cond_t*  end_cond;

		unsigned long int *sync, *end_sync;

		unsigned long int output_vector_size;

		unsigned long int output_vector_offset;

		T *local_y;

		const T ** input;

		T ** output;

		/** Base constructor. Will initialise all to invalid values or NULL. */
		RDB_shared_data(): id( -1 ), P( -1 ), mode( 0 ), repeat( 0 ), original( NULL ), nzb( NULL ), time( 0 ),
				mutex( NULL ), cond( NULL ), end_mutex( NULL ), end_cond( NULL ),
				sync( NULL ), end_sync( NULL ),
				output_vector_size( -1 ), output_vector_offset( -1 ),
				local_y( NULL ), input( NULL ), output( NULL ) {}

		/** Recommended constructor */
		RDB_shared_data( unsigned long int _id, unsigned long int _P,
				std::vector< Triplet< double > > *_original,
				unsigned long int *_nzb,
				pthread_mutex_t *_mutex, pthread_cond_t *_cond, pthread_mutex_t *_end_mutex, pthread_cond_t *_end_cond,
				unsigned long int *_sync, unsigned long int *_end_sync,
				unsigned long int _ovsize, unsigned long int _ovoffset,
				const T **_in, T **_out):
				id( _id ),  P( _P ), mode( 1 ), repeat( 1 ), original( _original ), nzb( _nzb ), time( 0 ),
				mutex( _mutex ), cond( _cond ), end_mutex( _end_mutex ), end_cond( _end_cond ),
				sync( _sync ), end_sync( _end_sync ),
				output_vector_size( _ovsize ), output_vector_offset( _ovoffset ),
				local_y( NULL ), input( _in ), output( _out ) {}
};

/** Compare function used for quicksort on an array of unsigned long ints */
int rdbh_uli_compare( const void *a, const void *b ) {
	return ( *(unsigned long int*)a - *(unsigned long int*)b );
}

/** The Beta Hilbert triplet scheme. Full parallel SpMV, based on Blocked Hilbert and PThreads.
 *  Inspired by Aydin & Gilbert's CSB, and comments by Patrick Amestoy on the BICRS Hilbert scheme. */
template< typename T >
class RDBHilbert: public SparseMatrix< T, ULI > {

	private:

	protected:

		/** Number of threads to fire up */
		static unsigned long int P;

		/** Which processors to pin threads to */
		std::vector< unsigned short int > p_translate;

		/** Input vector */
		const T* input;

		/** Output vector */
		T *output;

		/** p_threads associated to this data strcuture */
		pthread_t *threads;

		/** array of initial thread data */
		RDB_shared_data<T> *thread_data;

		/** Clock type used for thread-local timing */
		static clockid_t global_clock_id;

		/** Stop/continue mechanism: mutex */
		pthread_mutex_t mutex;

		/** Stop/continue mechanism: condition */
		pthread_cond_t cond;

		/** Wait for end mechanism: mutex */
		pthread_mutex_t end_mutex;

		/** Wait for end mechanism: condition */
		pthread_cond_t end_cond;

		/** Used for synchronising threads */
		unsigned long int sync;

		/** Used for construction end signal */
		unsigned long int end_sync;

		/** Given FBICRS, the maximum value for columnwise matrix size */
		static const ULI max_n = FBICRS< T >::beta_n;

		/** Given FBICRS, the maximum value for the rowwise matrix size, assuming short ints on ICRS at the lower level */
		static const ULI max_m = FBICRS< T >::beta_m;

		/** Sets p_translate to 0..P-1 by default, or equal to the optionally supplied vector. */
		void set_p_translate( std::vector< unsigned short int > *_p_translate ) {
			if( _p_translate == NULL ) {
				//get number of cores available
				P = MachineInfo::getInstance().cores();
				for( unsigned short int i = 0; i < P; ++i ) p_translate.push_back( i );
			} else {
				p_translate = *_p_translate;
				P = p_translate.size();
			}
		}

	public:

		RDBHilbert( const std::string file, T zero = 0, std::vector< unsigned short int > *_p_translate = NULL ) {
			set_p_translate( _p_translate );
			this->loadFromFile( file, zero );
		}

		RDBHilbert( std::vector< Triplet< T > >& input, ULI m, ULI n, T zero = 0, std::vector< unsigned short int > *_p_translate = NULL ) {
			set_p_translate( _p_translate );
			load( input, m, n, zero );
		}

		virtual ~RDBHilbert() {
			//set all daemon threads to exit mode
			for( unsigned long int i=0; i<P; i++ )
				thread_data[ i ].mode = 4;

			//wake up all daemon threads
			pthread_mutex_lock( &mutex );
			pthread_cond_broadcast( &cond );
			pthread_mutex_unlock( &mutex );

			//allow threads to exit gracefully
			for( unsigned long int i=0; i<P; i++ )
				pthread_join( threads[ i ], NULL );

			//destroy data
			delete [] thread_data;
			delete [] threads;
			pthread_mutex_destroy( &mutex );
			pthread_cond_destroy(  &cond  );
		}

		void wait() {
			//wait for end signal
			pthread_cond_wait( &end_cond, &end_mutex );
			pthread_mutex_unlock( &end_mutex );
		}

		virtual void load( std::vector< Triplet< T > >& input, const ULI m, const ULI n, const T zero ) {
			//set kernel to local thread allocation if it wasn't already the case
			numa_set_localalloc();

			//base settings
			this->zero_element = zero;
			this->nor = m;
			this->noc = n;
			this->nnz = input.size();

			// Parallel distribution according to the Hilbert curve:
			// -get maximum Hilbert coordinate
			//  (Note: curve coordinates should be implented such that (0,0) maps to 0!
			//         Otherwise a minimum Hilbert coordinate should also be found, and
			//         below code must be adapted to work on a Hilbert coordinate range
			//         instead)
			// -simultaneously also get number of nonzeroes to locally store
			// -translate to a range
			// -call function to distribute this range over P threads
			//
			// Most operations are done in parallel (see thread function), basically
			// only shared array allocation and the actual forking is in this function.

			unsigned long int *nzb = new unsigned long int [ this->m() ];
			for( unsigned long int i=0; i<m; i++ ) nzb[ i ] = 0;

			//create P threads :)
			this->threads = new pthread_t[ P ];
			//initialize local initialisation data
			thread_data = new RDB_shared_data<T>[ P ];
			//initialize mutexes and conditions and a synchronisation counter
			pthread_mutex_init( &mutex, NULL );
			pthread_cond_init ( &cond,  NULL );
			pthread_mutex_init( &end_mutex, NULL );
			pthread_cond_init ( &end_cond,  NULL );
			sync     = 0;
			end_sync = 0;
			//lock end mutex (disallow threads that are created to signal for end
			//before this thread is done with spawning children)
			pthread_mutex_lock( &end_mutex );
			//go forth and multiply
			for( unsigned long int i=0; i<P; i++ ) {
				//build thread-local init data
				thread_data[ i ] = RDB_shared_data<T>( i, P, &input, nzb, &mutex, &cond, &end_mutex, &end_cond, &sync, &end_sync, -1, -1, &(this->input), &output );
				//set fixed affinity for threads
				cpu_set_t mask;
				CPU_ZERO( &mask );
				CPU_SET ( p_translate[ i ], &mask );

				//TODO: use hwloc for better numa-aware pinning
				/*hwloc_topology_t topology;
				hwloc_topology_init ( &topology );
				hwloc_topology_load( topology );
				hwloc_bitmap_t cpuset;*/

				//prepare attributes
				pthread_attr_t attr;
				pthread_attr_init( &attr );
				//set fixed affinity in attribute, so that it starts binded immediately
				pthread_attr_setaffinity_np( &attr, sizeof( cpu_set_t ), &mask );
				//fire up thread
				pthread_create( &threads[i], &attr, &RDBHilbert::thread, (void*) &thread_data[i] );
				//free attr
				pthread_attr_destroy( &attr );
			}

			//wait for threads to finish initialisation
			wait();

			//delete temporary array
			delete [] nzb;
		}

		static void end( pthread_mutex_t* mutex, pthread_cond_t* cond, unsigned long int *sync, const unsigned long int P ) {
			pthread_mutex_lock( mutex );
			(*sync)++;
			if( *sync == P ) {
				pthread_cond_signal( cond );
				*sync = 0;
			}
			pthread_mutex_unlock( mutex );
		}

		static void synchronise( pthread_mutex_t* mutex, pthread_cond_t* cond, unsigned long int *sync, const unsigned long int P ) {
			pthread_mutex_lock( mutex );
			(*sync)++;
			if( *sync == P ) {
				*sync = 0;
				pthread_cond_broadcast( cond );
			} else
				pthread_cond_wait( cond, mutex );
			pthread_mutex_unlock( mutex );
		}

		static void* thread( void *data ) {
			//get short-hand notation
			RDB_shared_data<T>* shared  = (RDB_shared_data<T>*)data;
			const unsigned long int id  = shared->id;
			const unsigned long int P   = shared->P;
			const unsigned long int nnz = shared->original->size();
			pthread_mutex_t *mutex      = shared->mutex;
			pthread_cond_t  *cond       = shared->cond;

			//check if pinned to exactly one thread
			cpu_set_t mask;
			CPU_ZERO( &mask );
			pthread_getaffinity_np( pthread_self(), sizeof( cpu_set_t ), &mask );
			unsigned long int MyMask = ULONG_MAX;
			for( unsigned long int s=0; s<P; s++ ) {
				if( CPU_ISSET( s, &mask ) ) {
					if( MyMask == ULONG_MAX )
						MyMask = s;
					else {
						std::cerr << "Thread " << id << " mask is larger than one core" << " (" << MyMask << " and " << s << " are both set)!" << std::endl;
						exit( 1 );
					}
				}
			}

			//prepare to get global matrix dimensions
			ULI m, n;
			m = n = 0;
			//put rowsums in nzb
			const unsigned long int blocksize = (nnz % P) > 0 ? nnz / P + 1 : nnz / P;
			for( unsigned long int i=0; i<nnz; i++ ) {
				const unsigned long int currow = (*(shared->original))[ i ].i();
				const unsigned long int curcol = (*(shared->original))[ i ].j();
				if( currow >= id * blocksize && currow < (id + 1) * blocksize )
					shared->nzb[ currow ]++;
				if( currow > m ) m = currow;
				if( curcol > n ) n = curcol;
			}
			
			//dimensions are one higher than max indices
			m++;
			n++;

			//sync
			RDBHilbert::synchronise( mutex, cond, shared->sync, shared->P );

			//determine distribution
			const unsigned long int nnz_target = nnz / P;
			unsigned long int cursum = 0;

			//first sanity check
			for( unsigned long int i=0; i<m; i++ ) cursum += shared->nzb[ i ];
			assert( cursum == nnz );
			
			//continue
			cursum = 0;
			unsigned long int start, end, k = 0;
			start = end = -1;
			if( id == 0 ) start = 0;
			for( unsigned long int i=0; i<m; i++ ) {
				cursum += shared->nzb[ i ];	
				if( cursum >= nnz_target ) {
					if( k == id ) end   = i + 1;
					if(k+1== id ) start = i + 1;
					k++;
					cursum = 0;
				}
			}
			if( id == P-1 ) end = m;
#ifndef NDEBUG
			std::cout << "Thread " << id << " has rows " << start << " to " << end << std::endl;
#endif
			shared->output_vector_size   = end - start;
			shared->output_vector_offset = start;
			assert (shared->output_vector_size <= m );
			assert( shared->output_vector_offset + shared->output_vector_size <= m );

			//copy to local first			
			std::vector< Triplet< T > > local;
			for( unsigned long int i=0; i<nnz; i++ ) {
				const unsigned long int currow = (*(shared->original))[ i ].i();
				if( currow >= start && currow < end )
					local.push_back(
						Triplet< T >( (*(shared->original))[ i ].i() - start,
							(*(shared->original))[ i ].j(),
							(*(shared->original))[ i ].value )
					);
			}
			const unsigned long int local_nnz = local.size();
			m = shared->output_vector_size; //new matrix size is new m times old n

			//extract hilbert coordinate info
			unsigned long int h1, h2;
			Matrix2HilbertCoordinates::IntegerToHilbert( local[0].i() / max_m, local[0].j() / max_n, h1, h2 );
			BigInt maxh( h1, h2), minh( h1, h2 );
			BigInt *t2h = new BigInt[ local_nnz ];
			for( unsigned long int i=0; i<local_nnz; i++ ) {
				Matrix2HilbertCoordinates::IntegerToHilbert( local[i].i() / max_m, local[i].j() / max_n, h1, h2 );
				const BigInt cur( h1, h2 );
				t2h[ i ] = cur;
				if( maxh < cur ) maxh = cur;
				if( minh > cur ) minh = cur;			
				assert( t2h[ i ].high == h1 );
				assert( t2h[ i ].low  == h2 );
			}

			//sequential quicksort O(nz*log(nz))
			qsort(  t2h, local_nnz, sizeof( BigInt ), bigint_compare );
			assert( t2h[ local_nnz - 1 ] == maxh );
			assert( t2h[ 0 ] == minh );
				
			//build local data structure from the triplets in shared->original
			//and beta block hilbert index to beta translation map
			std::map< BigInt, unsigned long int > h2b;
			std::vector< std::vector< Triplet< T > > > beta;

			//prepare beta and map
			//get first hilbert block coordinate
			BigInt cur = t2h[ 0 ];
			//remember index
			h2b[ cur ] = 0;
			//keep next index
			unsigned long int c = 0, size = 1;
			//remember previous index
			BigInt prev_h = cur;
			//do the same for the remainder of available hilbert coordinates
			for( unsigned long int i = 1; i < local_nnz; ++i, ++size ) {
				cur = t2h[ i ];
				if( cur != prev_h ) { //new coordinate
					//push back old vector with exact size
					beta.push_back( std::vector< Triplet< T > >() );
					beta.back().reserve( size );
					//store previous index
					h2b[ prev_h ] = c++;
					//reset
					prev_h = cur;
					size = 1;
				}
			}
			//push back last one
			beta.push_back( std::vector< Triplet< T > >() );
			beta.back().reserve( size );
			h2b[ cur ] = c;
					
			//prepare to get matrix size (m, n) and some statistics too
			unsigned long int smin_m = ULONG_MAX;
			unsigned long int smin_n = ULONG_MAX;
			unsigned long int smax_m, smax_n;
			smax_m = smax_n = 0;
			unsigned long int *ms    = new unsigned long int[ h2b.size() ];
			unsigned long int *ns    = new unsigned long int[ h2b.size() ];
			unsigned long int *minms = new unsigned long int[ h2b.size() ];
			unsigned long int *minns = new unsigned long int[ h2b.size() ];
			for( unsigned long int i=0; i<h2b.size(); i++ ) {
				ms[ i ] = 0;
				ns[ i ] = 0;
				minms[ i ] = ULONG_MAX;
				minns[ i ] = ULONG_MAX;
			}

			//fill beta blocks in correct hilbert order O(local_nnz * log( nonzero_blocks ))
			for( unsigned long int i=0; i<local_nnz; i++ ) {
				const unsigned long int row = local[ i ].i();
				const unsigned long int col = local[ i ].j();
				Matrix2HilbertCoordinates::IntegerToHilbert( row / max_m, col / max_n, h1, h2 );
				const BigInt cur( h1, h2 );
				assert( cur >= minh );
				assert( cur <= maxh );
				assert( row <  m );
				const Triplet< T > toAdd = Triplet< T >( row, col, local[ i ].value );
#ifndef NDEBUG
				if( beta[ h2b[ cur ] ].size() > 1 ) {
					Matrix2HilbertCoordinates::IntegerToHilbert(
						beta[ h2b[ cur ] ].back().i() / max_m,
						beta[ h2b[ cur ] ].back().j() / max_n, h1, h2 );
					assert( cur == BigInt( h1, h2 ) );
					assert( beta[ h2b[ cur ] ].back().i() / max_m == row / max_m );
					assert( beta[ h2b[ cur ] ].back().j() / max_n == col / max_n );
				}
#endif
				beta[ h2b[ cur ] ].push_back( toAdd );
				if( row > ms[ h2b[ cur ] ] ) ms[ h2b[ cur ] ] = row;
				if( col > ns[ h2b[ cur ] ] ) ns[ h2b[ cur ] ] = col;
				if( row < minms[ h2b[ cur ] ] ) minms[ h2b[ cur ] ] = row;
				if( col < minns[ h2b[ cur ] ] ) minns[ h2b[ cur ] ] = col;
				if( row < smin_m ) smin_m = row;
				if( col < smin_n ) smin_n = col;
				if( row > smax_m ) smax_m = row;
				if( col > smax_n ) smax_n = col;
			}
			//size is max value + 1
			smax_m++; smax_n++;

#ifndef NDEBUG
			cursum = 0;
			for( unsigned long int i=0; i<beta.size(); i++ )
				cursum += beta[i].size();
			assert( cursum == local_nnz );
			std::cout << "Thread " << shared->id << ": " << smin_m << "," << smax_m << " times " << smin_n << "," << smax_n << " holding " << cursum << " nonzeroes." << std::endl;

			//sanity check: everything in one beta vector should share the
			//same block.
			for( unsigned long int i=0; i<beta.size(); ++i ) {
				const unsigned long int row_br = beta[i][0].i() / max_m;
				const unsigned long int col_br = beta[i][0].j() / max_n;
				for( unsigned long int k=1; k<beta[i].size(); ++k ) {
					assert( beta[i][k].i() / max_m == row_br );
					assert( beta[i][k].j() / max_n == col_br );
				}
			}
#endif

			//remove temporary values
			delete [] ms;
			delete [] ns;
			delete [] minms;
			delete [] minns;
			delete [] t2h;

			//load into FBICRS
			FBICRS< T > dss( beta, smax_m - smin_m, n );
			beta.clear();

			//create local shadow of y to avoid write-contention
			T* y = NULL;
#ifdef RDBH_GLOBAL_Y
			if( id > 0 ) {
#endif
				y = new T[ shared->output_vector_size ];
				for( unsigned long int i=0; i<shared->output_vector_size; i++ )
					y[ i ] = 0.0;
#ifdef RDBH_GLOBAL_Y
			}
#endif
			shared->local_y = y;
	
			//exit construction mode
			shared->mode = 0;

			//signal end of construction
			pthread_mutex_lock( mutex );
			RDBHilbert::end( shared->end_mutex, shared->end_cond, shared->end_sync, shared->P );

			//enter daemon mode
			while( true ) {
				struct timespec clk_start, clk_stop; 
				pthread_cond_wait(  cond, mutex );
				pthread_mutex_unlock( mutex );

				if( shared->mode == 4 ) break;

				switch( shared->mode ) {
				case 3:
					assert( *(shared->input) != NULL );
					assert( *(shared->output) != NULL );
#ifdef RDBH_GLOBAL_Y
					if( id == 0 ) {
						y = *(shared->output);
						shared->local_y = y;
					}
#endif
					assert( y != NULL );

					clock_gettime( global_clock_id, &clk_start);
					shared->time = 0.0;
					for( unsigned long int i=0; i<shared->repeat; i++ )
						dss.zxa( *(shared->input), y );
					clock_gettime( global_clock_id, &clk_stop);
					shared->time  = (clk_stop.tv_sec-clk_start.tv_sec)*1000;
					shared->time += (clk_stop.tv_nsec-clk_start.tv_nsec)/1000000.0;

#ifndef RDBH_NO_COLLECT
					collectY( shared );
#endif
					break;
				case 2:
					assert( *(shared->input) != NULL );
					assert( *(shared->output) != NULL );
#ifdef RDBH_GLOBAL_Y
					if( id == 0 ) {
						y = *(shared->output);
						shared->local_y = y;
					}
#endif
					assert( y != NULL );

					clock_gettime( global_clock_id, &clk_start);
					shared->time = 0.0;
					for( unsigned long int i=0; i<shared->repeat; i++ )

					dss.zax( *(shared->input), y );

					clock_gettime( global_clock_id, &clk_stop);
					shared->time  = (clk_stop.tv_sec-clk_start.tv_sec)*1000;
					shared->time += (clk_stop.tv_nsec-clk_start.tv_nsec)/1000000.0;

#ifndef RDBH_NO_COLLECT
					collectY( shared );
#endif
					break;
				default:
					std::cout << "Thread " << id << ": Error, undefined operation (" << shared->mode << ")!" << std::endl;
					exit( -1 );
				}
				shared->mode = 0;

				//signal end of operation
				pthread_mutex_lock( mutex );
				RDBHilbert::end( shared->end_mutex, shared->end_cond, shared->sync, shared->P );
			}

			//done
#ifdef RDBH_GLOBAL_Y
			if( id != 0 )
#endif
				delete [] y;
			return (NULL);
		}

		static void collectY( RDB_shared_data<T> *shared ) {

#ifdef RDBH_GLOBAL_Y
			//FIXME It could be possible to distribute work over all processors
			//instead of p-1 processors, but this requires some extra balancing.
			const unsigned long int s = shared->id;
			if( s == 0 ) return;
#endif

			//do collect items of own block
			for( unsigned long int i = 0; i < shared->output_vector_size; i++ ) {
				assert( *(shared->output) != NULL );
				assert( shared->local_y != NULL );
				(*(shared->output))[ shared->output_vector_offset + i ] += shared->local_y[ i ];
			}
		}

		/** Overloaded mv call; allocates output vector using numa_interleaved. */
		virtual T* mv( const T* x ) {
			T* ret = (T*) numa_alloc_interleaved( this->nor * sizeof( T ) );
			for( ULI i=0; i<this->nor; i++ ) ret[ i ] = this->zero_element;
			zax( x, ret );
			return ret;
		}

		virtual void zxa( const T* x, T* z ) {
			zxa( x, z, 1 );
		}

		virtual void zxa( const T* x, T* z, const unsigned long int repeat ) {
			//set all daemon threads to do zxa
			for( unsigned long int i=0; i<P; i++ ) {
				thread_data[ i ].mode   = 3;
				thread_data[ i ].repeat = repeat;
			}

			//set input vector
			input = x;

			//set output vector
			output = z;

			//wake up all daemon threads
			pthread_mutex_lock( &end_mutex );
			pthread_mutex_lock( &mutex );
			pthread_cond_broadcast( &cond );
			pthread_mutex_unlock( &mutex );

			//wait for end of operation
			wait();

			//unset vectors
			input  = NULL;
			output = NULL;
		}

		virtual void zax( const T* x, T* z ) {
			zax( x, z, 1, 0, NULL );
		}

		virtual void zax( const T* x, T* z, const unsigned long int repeat, const clockid_t clock_id, double *elapsed_time ) {
			//set all daemon threads to do zax
			for( unsigned long int i=0; i<P; i++ ) {
				thread_data[ i ].mode   = 2;
				thread_data[ i ].repeat = repeat;
			}

			//set global clock ID
			global_clock_id = clock_id;

			//set input vector
			input = x;

			//set output vector
			output = z;

			//wake up all daemon threads
			pthread_mutex_lock( &end_mutex );
			pthread_mutex_lock( &mutex );
			pthread_cond_broadcast( &cond );
			pthread_mutex_unlock( &mutex );

			//wait for end of operation
			wait();

			//get elapsed time
			double maxtime = 0.0;
			for( unsigned long int i=0; i<P; i++ ) {
				const double curtime = thread_data[ i ].time;
				if( curtime > maxtime ) maxtime = curtime;
			}
			if( elapsed_time != NULL )
				*elapsed_time += maxtime;

			//unset vectors
			input  = NULL;
			output = NULL;
		}

		virtual void getFirstIndexPair( ULI &i, ULI &j ) {
			std::cerr << "Warning: RDBHilbert::getFirstIndexPair has no unique answer since it implements a parallel multiplication!\nIgnoring call..." << std::endl;
		}
};

template< typename T > unsigned long int RDBHilbert< T >::P = 0;

template< typename T > clockid_t RDBHilbert< T >::global_clock_id = 0;

#endif

